#include <Adafruit_GPS.h>
#include <SPI.h>
#include <SD.h>

const int chipSelect = 10; // Pin for SD card chip select
const int buttonPin = 11;  // Pin connected to the button
const int ledPin = 13;     // Pin connected to an LED
char datalogFilename[20]; // Will hold something like "Datalog1.txt" 
char tracklogFilename[20]; // Will hold something like "Tracklog1.txt"

#define mySerial Serial1 // GPS serial port
Adafruit_GPS GPS(&mySerial);

#define GPSECHO false // Set to true to echo GPS data to Serial

unsigned long lastLogTime = 0;  // Track the last time data was logged
const unsigned long logInterval = 1000;  // 1 second interval for button press logging
const unsigned long tracklogInterval = 3000; // 3 second interval
unsigned long lastButtonPressTime = 0; // Track last button press time
unsigned long lastTracklogTime = 0; // Track last tracklog time

bool headerWrittenGPS = false;  // For GPS data
bool headerWrittenTracklog = false;  // For tracklog data

unsigned long buttonPressStartTime = 0; // Time when button press starts
bool buttonPressed = false; // Flag to check if button is pressed
bool sdCardInitialized = false; // Flag to track SD card initialization status

void setup() {
    Serial.begin(115200);
    delay(5000);
    Serial.println("Adafruit GPS logger starting up!");

    // Setup button pin and LED
    pinMode(buttonPin, INPUT_PULLUP); // Set button pin as input with pull-up resistor
    pinMode(ledPin, OUTPUT);           // Set LED pin as output
    
    // Signal the start of the initialization
    for (int i = 0; i < 3; i++) {
        digitalWrite(ledPin, HIGH);
        delay(200);
        digitalWrite(ledPin, LOW);
        delay(200);
    }

    // Initialize GPS
    Serial.println("Initializing GPS module...");
    GPS.begin(9600);
    GPS.sendCommand(PMTK_SET_NMEA_OUTPUT_RMCGGA);
    GPS.sendCommand(PMTK_SET_NMEA_UPDATE_1HZ); // 1 Hz update rate
    GPS.sendCommand(PGCMD_ANTENNA);
    delay(1000);
    mySerial.println(PMTK_Q_RELEASE);
    Serial.println("GPS initialization complete");

    // Initialize SD card with more robust error handling
    initializeSDCard();
    
    // Reset the tracklog timer
    lastTracklogTime = millis();
}

void loop() {
    char c = GPS.read();
    if (c && GPSECHO) {
        Serial.write(c);
    }

    if (GPS.newNMEAreceived()) {
        String nmea = GPS.lastNMEA();
        if (!GPS.parse(nmea.c_str())) {  // Use c_str() to convert to C-style string
            return;
        }
    }

    // Only proceed with logging if SD card is initialized
    if (sdCardInitialized) {
        // Check if button is pressed for main log
        if (digitalRead(buttonPin) == LOW) {
            if (!buttonPressed) {
                buttonPressed = true;
                buttonPressStartTime = millis(); // Mark the start of the button press
            } else {
                // If the button is pressed, check if enough time has passed for a new log
                if (millis() - lastButtonPressTime >= logInterval) {
                    lastButtonPressTime = millis(); // Update last log time
                    digitalWrite(ledPin, HIGH); // Turn LED on while logging
                    recordGPSData(); // Log the GPS data
                    digitalWrite(ledPin, LOW); // Turn LED off after logging
                }
            }
        } else {
            // Button is released
            if (buttonPressed) {
                buttonPressed = false;
            }
        }

        // Tracklog every 3 seconds
        if (millis() - lastTracklogTime >= tracklogInterval) {
            lastTracklogTime = millis(); // Update the last tracklog time
            digitalWrite(ledPin, HIGH); // Turn LED on while logging
            recordTracklogData(); // Log track data
            delay(100); // Keep LED on briefly to be visible
            digitalWrite(ledPin, LOW); // Turn LED off after logging
        }
    } else {
        // If SD card not initialized, try again every 30 seconds
        static unsigned long lastSDInitAttempt = 0;
        if (millis() - lastSDInitAttempt >= 30000) {
            lastSDInitAttempt = millis();
            Serial.println("Retrying SD card initialization...");
            initializeSDCard();
        }
        
        // Blink LED to indicate SD card error
        static unsigned long lastBlink = 0;
        if (millis() - lastBlink >= 500) {
            lastBlink = millis();
            static bool ledState = false;
            ledState = !ledState;
            digitalWrite(ledPin, ledState);
        }
    }
}

// Function to initialize SD card with better error handling
void initializeSDCard() {
    Serial.print("Initializing SD card...");
    
    // Power cycle the SD card if possible (via software)
    pinMode(chipSelect, OUTPUT);
    digitalWrite(chipSelect, HIGH);
    delay(250);
    
    // Try to initialize the SD card
    if (!SD.begin(chipSelect)) {
        Serial.println("Initialization failed. Check:");
        Serial.println("1. Is a card inserted?");
        Serial.println("2. Is your wiring correct?");
        Serial.println("3. Did you change the chipSelect pin?");
        Serial.println("4. Is the card formatted as FAT16/FAT32?");
        sdCardInitialized = false;
        return;
    }
    
    // Card type detection - if available on your SD library version
    Serial.println("SD Card initialized successfully!");
    
    // Test writing a simple file to verify the card works
    File testFile = SD.open("test.txt", FILE_WRITE);
    if (testFile) {
        testFile.println("SD card test successful");
        testFile.close();
        Serial.println("Test file created successfully");
        sdCardInitialized = true;
    } else {
        Serial.println("Error creating test file - card may be write-protected");
        sdCardInitialized = false;
        return;
    }
    
    // If we got here, the card is working properly
    Serial.println("SD Card verified working properly!");
    
    // Generate new filenames for this session
    if (generateFilenames()) {
        Serial.println("Ready to log GPS data!");
    } else {
        Serial.println("Failed to create log files");
        sdCardInitialized = false;
    }
}

// Function to generate sequential filenames
bool generateFilenames() {
    int datalogNum = 1;
    int tracklogNum = 1;
    
    // Find the next available datalog filename
    while (true) {
        sprintf(datalogFilename, "DATALOG%d.TXT", datalogNum);
        if (!SD.exists(datalogFilename)) {
            break;
        }
        datalogNum++;
    }
    
    // Find the next available tracklog filename
    while (true) {
        sprintf(tracklogFilename, "TRAKLOG%d.TXT", tracklogNum);
        if (!SD.exists(tracklogFilename)) {
            break;
        }
        tracklogNum++;
    }
    
    Serial.print("New datalog file: ");
    Serial.println(datalogFilename);
    Serial.print("New tracklog file: ");
    Serial.println(tracklogFilename);
    
    // Create both files with headers
    bool datalogSuccess = createFile(datalogFilename);
    bool tracklogSuccess = createFile(tracklogFilename);
    
    return datalogSuccess && tracklogSuccess;
}

// Helper function to create a file and write the header
bool createFile(const char* filename) {
    // First try to create the file
    File file = SD.open(filename, FILE_WRITE);
    if (!file) {
        Serial.print("Error opening file for creation: ");
        Serial.println(filename);
        return false;
    }
    
    // Write header
    file.println("Date,Time_UTC,Latitude,Longitude,Elevation,Speed(m/s),FixType");
    file.close();
    Serial.print("Successfully created and initialized ");
    Serial.println(filename);
    return true;
}

void recordData(File& file, bool isTracklog) {
    if (!file) {
        Serial.println("Error opening file for recording data");
        return;
    }

    if (GPS.fix) {
        // Print Date first (YYYY-MM-DD)
        file.print(GPS.year + 2000); // Print the year as YYYY
        file.print('-');
        file.print(GPS.month < 10 ? "0" : "");
        file.print(GPS.month, DEC);
        file.print('-');
        file.print(GPS.day < 10 ? "0" : "");
        file.print(GPS.day, DEC);
        file.print(",");  // Comma separator

        // Print Time after Date (HH:MM:SS)
        file.print(formatTime(GPS.hour, GPS.minute, GPS.seconds));
        file.print(",");

        // Latitude, Longitude, Elevation, Speed (in m/s), Fix Type
        file.print(convertToDecimal(GPS.latitude, GPS.lat == 'S'), 6);
        file.print(",");
        file.print(convertToDecimal(GPS.longitude, GPS.lon == 'W'), 6);
        file.print(",");
        file.print(GPS.altitude);
        file.print(",");

        float speedMps = GPS.speed * 0.514444; // Convert knots to m/s
        file.print(speedMps, 2);
        file.print(",");
        
        file.print(GPS.fix ? (GPS.fixquality == 0 ? "No Fix" : (GPS.fixquality == 1 ? "2D Fix" : "3D Fix")) : "No Fix");

        file.println();
        
        if (isTracklog) {
            Serial.println("Tracklog data recorded successfully.");
        } else {
            Serial.println("Datalog data recorded successfully.");
        }
    } else {
        Serial.println("GPS not fixed, no data logged.");
    }
}

void recordGPSData() {
    File dataFile = SD.open(datalogFilename, FILE_WRITE);
    if (dataFile) {
        recordData(dataFile, false);
        dataFile.close();
    } else {
        Serial.print("Error opening datalog file: ");
        Serial.println(datalogFilename);
        
        // If we fail to open the file, the SD card might be having issues
        // Try reinitializing the SD card
        sdCardInitialized = false;
    }
}

void recordTracklogData() {
    File tracklogFile = SD.open(tracklogFilename, FILE_WRITE);
    if (tracklogFile) {
        recordData(tracklogFile, true);
        tracklogFile.close();
    } else {
        Serial.print("Error opening tracklog file: ");
        Serial.println(tracklogFilename);
        
        // If we fail to open the file, the SD card might be having issues
        // Try reinitializing the SD card
        sdCardInitialized = false;
    }
}

// Format the time into a string (HH:MM:SS)
String formatTime(int hour, int minute, int seconds) {
    String timeStr = "";
    timeStr += (hour < 10 ? "0" : "") + String(hour) + ':';
    timeStr += (minute < 10 ? "0" : "") + String(minute) + ':';
    timeStr += (seconds < 10 ? "0" : "") + String(seconds);
    return timeStr;
}

// Convert GPS coordinates to decimal
float convertToDecimal(float coord, bool isNegative) {
    float degrees = (int)(coord / 100);
    float minutes = coord - (degrees * 100);
    float decimal = degrees + (minutes / 60.0);
    return isNegative ? -decimal : decimal;
}